'
'  Spin Code for the QMC5883L compass module.
'  Note: I got this code working without pull-ups.  This code writes x,y,z,and azimuth data to the PST.
'  Many thanks to JonnyMac, as some of the methods were his.

OBJ

    pst       :   "FullDuplexSerial"

CON

  _clkmode      = xtal1 + pll16x
  _clkfreq      = 80_000_000

  datapin  = 17         ' SDA of compass
  clockPin = 16         ' SCL of compass

con

  #0, ACK, NAK

con { object }

  QMC5883_WR = %0001_101_0
  QMC5883_RD = %0001_101_1

con { QMC5883L regs }

  R_X_LSB  = $00
  R_X_MSB  = $01
  R_Y_LSB  = $02
  R_Y_MSB  = $03
  R_Z_LSB  = $04
  R_Z_MSB  = $05

  R_CTRL1  = $09
  R_CTRL2  = $0A
  R_SETRST = $0B

  R_DEVID  = $0D

VAR

  long x
  long y
  long z
  long  scl                                                     ' bus pins
  long  sda

PUB Main

  waitcnt(clkfreq/100_000 + cnt)
  pst.start(31, 30, 0, 9600)
  init(clockpin, datapin)                                             ' start QMC5883

  repeat
    read_xyz
    pst.tx(1)
    DEG
    pst.tx(99)

pub init(sclpin, sdapin) : result

  setupx(sclpin, sdapin)
  waitcnt(cnt + (clkfreq / 20))

  if (present)
    write8(R_SETRST, $01)
    write8(R_CTRL1,  $11)                                       ' continuous, 10Hz, 8g, os512
    return true

  else
    return false

pub setupx(sclpin, sdapin)

'' Define I2C SCL (clock) and SDA (data) pins

  longmove(@scl, @sclpin, 2)                                    '  copy pins
  dira[scl] := 0                                                '  float to pull-up
  outa[scl] := 0                                                '  write 0 to output reg
  dira[sda] := 0
  outa[sda] := 0

  repeat 9                                                      ' reset bus
    if (ina[sda])
      quit
    dira[scl] := 1
    dira[scl] := 0

pub start

'' Create I2C (re)start sequence

  dira[sda] := 0                                                ' SDA high (float to pull-up)
  dira[scl] := 0                                                ' SCL high
  dira[scl] := 0                                                ' short delay

  dira[sda] := 1                                                ' SDA low
  dira[scl] := 1                                                ' SCL low

pub write(i2cbyte) : ackbit

'' Write byte to I2C bus
'' -- leaves SCL low

  i2cbyte := !i2cbyte << 24                                     ' prep for msbfirst to dir[]

  repeat 8                                                      ' 8 bits
    dira[sda] := i2cbyte <-= 1                                  ' msb first
    dira[scl] := 0                                              ' SCL high
    repeat while (ina[scl] == 0)                                ' hold (allow clock stretch)
    dira[scl] := 1                                              ' SCL low

  dira[sda] := 0                                                ' relase SDA
  dira[scl] := 0                                                ' SCL high
  repeat while (ina[scl] == 0)                                  ' hold (allow clock stretch)
  ackbit := ina[sda]                                            ' read ack bit
  dira[scl] := 1                                                ' SCL low

pub read(ackbit) : i2cbyte

'' Read byte from I2C bus

  dira[sda] := 0                                                ' release SDA

  repeat 8
    dira[scl] := 0                                              ' SCL high
    repeat while (ina[scl] == 0)                                ' hold (allow clock stretch)
    i2cbyte := (i2cbyte << 1) | ina[sda]                        ' read the bit
    dira[scl] := 1                                              ' SCL low

  dira[sda] := !ackbit                                          ' output ack bit
  dira[scl] := 0                                                ' SCL high
  repeat while (ina[scl] == 0)                                  ' hold (allow clock stretch)
  dira[scl] := 1                                                ' SCL low

pub stop

'' Create I2C stop sequence

  dira[sda] := 1                                                ' SDA low
  dira[sda] := 1                                                ' delay

  dira[scl] := 0                                                ' SCL high
  dira[sda] := 0                                                ' SDA high

pri write8(reg, value)

  start
  write(QMC5883_WR)
  write(reg)
  write(value)
  stop

pri read8(reg) : value

  start
  write(QMC5883_WR)
  write(reg)
  start
  write(QMC5883_RD)
  result := read(NAK)
  stop

pub present : result

  start
  result := write(QMC5883_WR)
  stop

  return (result == ACK)

pub read_xyz | tx, ty, tz

  longfill(@tx, 0, 3)                                           ' clear temp registers

  start                                                     ' read signed 16-bit values
  write(QMC5883_WR)
  write(R_X_LSB)
  start
  write(QMC5883_RD)
  tx.byte[0] := read(ACK)
  tx.byte[1] := read(ACK)
  ty.byte[0] := read(ACK)
  ty.byte[1] := read(ACK)
  tz.byte[0] := read(ACK)
  tz.byte[1] := read(NAK)
  stop

  x := ~~tx                                                 ' convert to longs
  y := ~~ty
  z := ~~tz

PUB DEG | value

    if  x =< 0
        if azimuth =< 0
           value := AZ_A

        if azimuth > 0
           value := AZ_D

    if x > 0
        if azimuth =< 0
          value := AZ_B

        if azimuth > 0
          value := AZ_C

  value := 1 #> value <# 360

  value := (value + 270) // 360

  PST.str(string(11, 13, " Heading "))
  PST.dec(value)
  PST.str(string(11, 13, " X "))
  PST.dec(x)
  PST.str(string(11, 13, " Y "))
  PST.dec(y)
  PST.str(string(11, 13, " Z "))
  PST.dec(z)

PRI Azimuth

  result := arctan(y,x)

PRI AZ_A        'NE

               result := || azimuth

PRI AZ_B        'SE

               result := azimuth  + 180

PRI AZ_C        'SW

               result := azimuth  + 180

PRI AZ_D        'NW

               result := - azimuth + 360

PUB arcSin(yLength, radius) | mirrorFlag, minValue, maxValue '' 8 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Computes the arc sin of an angle.
'' //
'' // Returns the angle in which ((radius) * sin(angle)) = yLength.
'' //
'' // yLength - The opposite side of a right triangle. Opposite to where the angle is calculated.
'' // Radius - The radius (hypotenuse) of the polar plane triangle.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  radius := ((radius <# 32_767) #> 0)
  yLength := ((yLength <# radius) #> -radius)

  mirrorFlag := (yLength < 0)
  yLength := (((||yLength) * $FF_FF) / radius)

  minValue := 0
  maxValue := 20_48

  repeat

    result := ((minValue + maxValue) >> 1)

    if(yLength => word[$E0_00][result++])

      minValue := result

    else

      maxValue := (result - 2)

  until(minValue > maxValue)

  return (((result * 90) >> 11) * (mirrorFlag | 1))

PUB arcTan(yLength, xLength) '' 13 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Computes the arc tan of an angle.
'' //
'' // Returns the angle in which tan(angle) = (yLength / xLength).
'' //
'' // yLength - The opposite side of a right triangle. Opposite to where the angle is calculated.
'' // xLength - The addjacent side of a right triangle. Addjacent to where the angle is calculated.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return arcSin(yLength, polarRadius(yLength, xLength))

PUB polarRadius(yLength, xLength) '' 5 Stack Longs

'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
'' // Computes the radius (hypotenuse) of a triangle in the polar plane using its x and y side lengths (pythagorian theram).
'' //
'' // Returns the radius (hypotenuse) of a triangle in the polar plane.
'' //
'' // yLength - The opposite side of a right triangle. Opposite to where the angle is calculated.
'' // xLength - The addjacent side of a right triangle. Addjacent to where the angle is calculated.
'' ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  return (^^((xLength * xLength) + (yLength * yLength)))